"""
=================================
Tools (:mod:`CADETProcess.tools`)
=================================

.. currentmodule:: CADETProcess.tools

A collection of tools extending the functionatlity of **CADET-Process**.

Yamamoto's Method
=================

A module to fit isotherm parameters using Yamamoto's method.

.. autosummary::
   :toctree: generated/

   yamamoto

"""
