"""
=================================================
Fractionation (:mod:`CADETProcess.fractionation`)
=================================================

.. currentmodule:: CADETProcess.fractionation

A module to calculate process performance indicators.

.. autosummary::
   :toctree: generated/

   Fraction
   FractionPool
   Fractionator
   FractionationOptimizer

"""

from .fractions import *
from .fractionator import *
from .fractionationOptimizer import *
