"""
=========================================================
CADETProcessError (:mod:`CADETProcess.CADETProcessError`)
=========================================================

.. currentmodule:: CADETProcess.CADETProcessError

Exceptions in **CADET-Process**.

.. autosummary::
    :toctree: generated/

    CADETProcessError

"""  # noqa


class CADETProcessError(Exception):
    """Exception for errors in CADET-Process."""
