"""
==================================================
Dynamic Events (:mod:`CADETProcess.dynamicEvents`)
==================================================

.. currentmodule:: CADETProcess.dynamicEvents

This module provides functionality for dynamic changes of parameters in CADET-Process.

Sections and Timelines
======================

.. autosummary::
    :toctree: generated/

    Section
    TimeLine
    MultiTimeLine

Events and EventHandlers
========================

.. autosummary::
    :toctree: generated/

    Event
    Duration
    EventHandler

"""

from .section import *
from .event import *
